@@ -27,6 +27,7 @@
 #include <string.h>
 #include <glib/gi18n-lib.h>
 #include <gtk/gtk.h>
+#include <libhal.h>
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
 #include <eel/eel-features.h>
@@ -62,6 +63,7 @@ struct _NautilusBurnPrivate
 	guint                  empty : 1;
 
 	guint                  start_monitor_id;
+	guint                  empty_update_id;
 
 	GSList                *widget_list;
 };
@@ -258,6 +260,70 @@ copy_disc_activate_cb (NautilusMenuItem *item,
 	g_free (argv [0]);
 }
 
+
+static LibHalContext *
+get_hal_context (void)
+{
+	static LibHalContext *ctx = NULL;
+	DBusError	      error;
+	DBusConnection	     *dbus_conn;
+
+	if (ctx == NULL) {
+		ctx = libhal_ctx_new ();
+		if (ctx == NULL) {
+			g_warning ("Could not create a HAL context");
+		} else {
+			dbus_error_init (&error);
+			dbus_conn = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
+
+			if (dbus_error_is_set (&error)) {
+				g_warning ("Could not connect to system bus: %s", error.message);
+				dbus_error_free (&error);
+				return NULL;
+			}
+
+			libhal_ctx_set_dbus_connection (ctx, dbus_conn);
+
+			if (! libhal_ctx_init (ctx, &error)) {
+				g_warning ("Could not initalize the HAL context: %s",
+					   error.message);
+
+				if (dbus_error_is_set (&error))
+					dbus_error_free (&error);
+
+				libhal_ctx_free (ctx);
+				ctx = NULL;
+			}
+		}
+	}
+
+	return ctx;
+}
+
+static gboolean
+volume_is_blank (GnomeVFSVolume *volume)
+{
+	LibHalContext *ctx;
+	DBusError      error;
+	char          *udi;
+	gboolean       is_blank;
+
+	ctx = get_hal_context ();
+	if (ctx == NULL) {
+		return FALSE;
+	}
+
+	udi = gnome_vfs_volume_get_hal_udi (volume);
+	dbus_error_init (&error);
+	is_blank = libhal_device_get_property_bool (ctx,
+						    udi,
+						    "volume.disc.is_blank",
+						    NULL);
+	g_free (udi);
+
+	return is_blank;
+}
+
 static GList *
 nautilus_burn_get_file_items (NautilusMenuProvider *provider,
 			      GtkWidget            *window,
@@ -269,7 +335,6 @@ nautilus_burn_get_file_items (NautilusMenuProvider *provider,
 	GnomeVFSFileInfo *info;
 	GnomeVFSVolume   *volume;
 	GnomeVFSDrive    *drive;
-	char             *device_path;
 	char             *mime_type;
 	gboolean          is_local;
 	gboolean          is_iso;
@@ -301,13 +366,14 @@ nautilus_burn_get_file_items (NautilusMenuProvider *provider,
 
 	is_iso = (strcmp (mime_type, "application/x-iso-image") == 0)
 		|| (strcmp (mime_type, "application/x-cd-image") == 0);
-	is_cue          = strcmp (mime_type, "application/x-cue") == 0;
+	is_cue = strcmp (mime_type, "application/x-cue") == 0
+		|| (strcmp (mime_type, "application/x-cdrdao-toc") == 0);
 
 	if (is_iso && is_local) {
 		item = nautilus_menu_item_new ("NautilusBurn::write_iso",
 					       _("_Write to Disc..."),
 					       _("Write disc image to a CD or DVD disc"),
-					       "gnome-dev-cdrom");
+					       "nautilus-cd-burner");
 		g_object_set_data (G_OBJECT (item), "file", file);
 		g_object_set_data (G_OBJECT (item), "window", window);
 		g_signal_connect (item, "activate",
@@ -317,7 +383,7 @@ nautilus_burn_get_file_items (NautilusMenuProvider *provider,
 		item = nautilus_menu_item_new ("NautilusBurn::write_cue",
 					       _("_Write to Disc..."),
 					       _("Write disc image cuesheet to a CD or DVD disc"),
-					       "gnome-dev-cdrom");
+					       "nautilus-cd-burner");
 		g_object_set_data (G_OBJECT (item), "file", file);
 		g_object_set_data (G_OBJECT (item), "window", window);
 		g_signal_connect (item, "activate",
@@ -326,22 +392,25 @@ nautilus_burn_get_file_items (NautilusMenuProvider *provider,
 	}
 
 	drive = nautilus_file_info_get_drive (file);
-	if (drive == NULL) {
-		volume = nautilus_file_info_get_volume (file);
-		if (volume != NULL) {
-			drive = gnome_vfs_volume_get_drive (volume);
-			gnome_vfs_drive_unref (drive);
-		}
+	volume = nautilus_file_info_get_volume (file);
+	if (drive == NULL
+	    && volume != NULL) {
+		drive = gnome_vfs_volume_get_drive (volume);
+		gnome_vfs_drive_unref (drive);
 	}
 
 	if (drive != NULL
-	    && (gnome_vfs_drive_get_device_type (drive) == GNOME_VFS_DEVICE_TYPE_CDROM)) {
+	    && volume != NULL
+	    && (gnome_vfs_drive_get_device_type (drive) == GNOME_VFS_DEVICE_TYPE_CDROM)
+	    && !volume_is_blank (volume)) {
+		char *device_path;
+
 		device_path = gnome_vfs_drive_get_device_path (drive);
 
 		item = nautilus_menu_item_new ("NautilusBurn::copy_disc",
 					       _("_Copy Disc..."),
 					       _("Create a copy of this CD or DVD disc"),
-					       "gnome-dev-cdrom");
+					       "nautilus-cd-burner");
 		g_object_set_data (G_OBJECT (item), "file", file);
 		g_object_set_data (G_OBJECT (item), "window", window);
 		g_object_set_data_full (G_OBJECT (item), "drive_device_path", device_path, g_free);
@@ -373,10 +442,12 @@ nautilus_burn_get_background_items (NautilusMenuProvider *provider,
 		item = nautilus_menu_item_new ("NautilusBurn::write_menu",
 					       _("_Write to Disc..."),
 					       _("Write contents to a CD or DVD disc"),
-					       "gnome-dev-cdrom");
+					       "nautilus-cd-burner");
 		g_signal_connect (item, "activate",
 				  G_CALLBACK (write_activate_cb), window);
 		items = g_list_append (items, item);
+
+		g_object_set (item, "sensitive", ! NAUTILUS_BURN (provider)->priv->empty, NULL);
 	}
 
 	g_free (scheme);
@@ -439,7 +510,8 @@ dir_is_empty (const char *uri)
 		}
 
 		/* Skip "." and "..".  */
-		if (info->name [0] == '.'
+		if (info->name != NULL
+		    && info->name [0] == '.'
 		    && (info->name [1] == 0
 			|| (info->name [1] == '.' && info->name [2] == 0))) {
 			gnome_vfs_file_info_clear (info);
@@ -518,6 +590,39 @@ update_widget_sensitivity (GtkWidget    *widget,
 	gtk_widget_set_sensitive (widget, !burn->priv->empty);
 }
 
+static gboolean
+update_empty_idle (NautilusBurn *burn)
+{
+	gboolean is_empty;
+
+	burn->priv->empty_update_id = 0;
+
+	is_empty = dir_is_empty ("burn:///");
+
+	DEBUG_PRINT ("Dir is %s\n", is_empty ? "empty" : "not empty");
+
+	if (burn->priv->empty != is_empty) {
+		burn->priv->empty = is_empty;
+		/* update bar */
+		g_slist_foreach (burn->priv->widget_list, (GFunc)update_widget_sensitivity, burn);
+
+		/* Trigger update for menu items */
+		nautilus_menu_provider_emit_items_updated_signal (NAUTILUS_MENU_PROVIDER (burn));
+	}
+
+	return FALSE;
+}
+
+static void
+queue_update_empty (NautilusBurn *burn)
+{
+	if (burn->priv->empty_update_id != 0) {
+		g_source_remove (burn->priv->empty_update_id);
+	}
+
+	burn->priv->empty_update_id = g_idle_add ((GSourceFunc)update_empty_idle, burn);
+}
+
 static void
 burn_monitor_cb (GnomeVFSMonitorHandle    *handle,
 		 const gchar              *text_uri,
@@ -526,21 +631,20 @@ burn_monitor_cb (GnomeVFSMonitorHandle    *handle,
 		 gpointer                  data)
 {
 	NautilusBurn *burn;
-	gboolean      is_empty;
 
 	burn = NAUTILUS_BURN (data);
 
-	is_empty = burn->priv->empty;
+	DEBUG_PRINT ("Monitor callback type %d: %s: %s\n", event_type, text_uri, info_uri);
 
+	/* only queue the action if it has a chance of changing the state */
 	if (event_type == GNOME_VFS_MONITOR_EVENT_CREATED) {
-		is_empty = FALSE;
+		if (burn->priv->empty) {
+			queue_update_empty (burn);
+		}
 	} else if (event_type == GNOME_VFS_MONITOR_EVENT_DELETED) {
-		is_empty = dir_is_empty ("burn:///");
-	}
-
-	if (burn->priv->empty != is_empty) {
-		burn->priv->empty = is_empty;
-		g_slist_foreach (burn->priv->widget_list, (GFunc)update_widget_sensitivity, burn);
+		if (! burn->priv->empty) {
+			queue_update_empty (burn);
+		}
 	}
 }
 
@@ -606,6 +710,10 @@ nautilus_burn_finalize (GObject *object)
 
 	g_return_if_fail (burn->priv != NULL);
 
+	if (burn->priv->empty_update_id > 0) {
+		g_source_remove (burn->priv->empty_update_id);
+	}
+
 	if (burn->priv->start_monitor_id > 0) {
 		g_source_remove (burn->priv->start_monitor_id);
 	}
@@ -693,6 +801,11 @@ nautilus_module_initialize (GTypeModule *module)
 void
 nautilus_module_shutdown (void)
 {
+	LibHalContext *ctx;
+	ctx = get_hal_context ();
+	if (ctx != NULL) {
+		libhal_ctx_free (ctx);
+	}
 }
 
 void
